% o  o o o o o     
%    o o o o o                                           
%    o o o o o    VariSphear VSA                            
%    o o o o o    Variable Spherical Microphone Array  
%    o o o                                            
%                                                    
%    /// ASAR RESEARCH GROUP          
%
%    Cologne University of Applied Sciences         
%    Deutsche Telekom Laboratories, Berlin          
%    Technical University of Berlin                 
%    Westdeutscher Rundfunk, Cologne         
%    
%
%
% VariSphear DRV - Device Drivers : Version R11-0520
% -----------------------------------------------------------------------
%
%
% Communication gateway to the motors of the VariSphear microphone 
% array unit and the modified KEMAR Head at T-Labs, Berlin.  
%
% Device Type: Schunk PR-70 Motion Module (AMTEC ROBOTICS)
% 
% /// Feel free to modify and share BUT: Always mark your changes please!
%
%
% motorreturn = vs_motion(COMPORT, DEVICE, TODO, POS, SPD, ACC)
%
% COMPORT : COM Port, e.g. 'COM5', Type: String 
% 
% DEVICE  : Motion Device, Type: String 
%
%      'az'    - Azimuth   (Ground Plane)   Range:   0 to +360
%      'el'    - Elevation (Boom)           Range:   0 to +180
%      'kemar' - Kemar Head T-Labs          Range: -90 to +90
%
% TODO    : Motor Command, Type: String
%
%    'mov' - Move to position in Degree RANGE: 0 to +360 DEG
%            The motor only accepts angles in the given range. 
%            Do a modulo operation if required.
%            'mov' is non-blocking, so you can use the motion time
%            for calculations or other actions.   
%    'stp' - Stop
%    'inf' - Status Information (real position, motor status flags etc.)
%    'cal' - Calibrate motor  
%            !!! WARNING: Always drive to 0 before sending a 'cal' cmd!
%            AZ and KEMAR calibrate to 0
%            EL calibrates to 90   
%    'err' - Quit internal Error   
% 
% SPD     : (SPEED) Between 1-150/s          #Default:50
% ACC     : (ACCELERATION) Between 1-250/s  #Default:50
%
%
% ----------------------------------------------------------------------
% 
% motorreturn:  returns an information struct (depending on the command)
%               motorreturn.flag always returns 0 if everything is ok.
%               
%               Error-Flags:
%               
%                0 : everything fine
%               -1 : arguments missing
%               -2 : config file missing 
%               -3 : device unknown
%               -4 : command unknown
%               -5 : motor returns error
%               -6 : communication error 
%
% ----------------------------------------------------------------------
%
% Advice :         - WARNING: ALWAYS drive the motor to the 0 position
%                             BEFORE calibrating the motor.
%                  - Look up the COM-Port for your specific system ->  
%                    NPort Driver Software or WIN Hardware Preferences 
%
% Dependencies :   VSAdeviceDriver.mexw32   (WIN32)
%                  VSAdeviceDriver.mexw64   (WIN64)
%                  motionConfig.mat         (Config File)
%                  crcTable.mat             (Checksum)
%
% Author        :  bBrn - Benny Bernschtz   
% Version       :  20M5.11 
% Last Version  :  V1.0 VariSphear FFL
% Since V1.0    :  - New lowlevel device driver
%                  - Improved Information Struct
%                  - WARNING: SYNTAX IS NOT COMPATIBLE TO FFL v1.0


function motorreturn=vs_motion(COMPORT, DEVICE, TODO, POS, SPD, ACC)

if nargin<3
    motorreturn = struct('status','FAILED. Not enough arguments',...
                         'flag',{-1});
    return

elseif nargin==3 
       POS=0;
       SPD=50;
       ACC=50;     
elseif nargin==4      
       SPD=50;
       ACC=50; 
elseif nargin==5      
       ACC=50;
end

try
    load motionConfig.mat deviceconfig commands
catch
    motorreturn.status = 'FAILED. Config file missing.';
    motorreturn.flag   = -2;
    return
end

deviceid = sum(strcmp(deviceconfig.device, DEVICE).*(1:length(deviceconfig.device)));

if ~deviceid
    motorreturn.status = ['FAILED. Device (', DEVICE ,') unknown.'];
    motorreturn.flag   = -3;
    return
end

command = sum(strcmp(commands.call, TODO).*(1:length(commands.call)));

if ~command
    motorreturn.status = ['FAILED. Command (', TODO ,') unknown.'];
    motorreturn.flag   = -4;
    return
end


motorreturn.device  = deviceconfig.name{deviceid};
motorreturn.command = commands.name{command};

COMPORT = upper(COMPORT);
comNumber           = str2double(strrep(COMPORT,'COM',''));


if command==1 
    limits = deviceconfig.limits{deviceid};   
    
    if (POS < limits(1) || POS > limits(2))
       motorreturn = struct('status',['FAILED. Angle out of range. Min: ', num2str(limits(1)), ', Max: ',num2str(limits(2)), ''], 'flag',{-1});
       return
    end
    
    if  deviceconfig.invert{deviceid}
        POS = deviceconfig.offset{deviceid}-POS;
    end
    
        posieee  = dec2ieee(POS);
        speedhex = dec2ieee(SPD);
        accelhex = dec2ieee(ACC);
        
        motor_cmd = hex2dec(crc16motor([commands.cmd{command}; posieee(1,:); posieee(2,:); posieee(3,:); posieee(4,:);...
                          speedhex(1,:); speedhex(2,:); speedhex(3,:); speedhex(4,:);...
                          accelhex(1,:); accelhex(2,:); accelhex(3,:); accelhex(4,:)]));  
    
else
   motor_cmd = hex2dec(crc16motor(commands.cmd{command}));
end

cmdRetry=0;
while(1)

    comRetry=0;    
     while(1)         
       [motorFlag, motorResponse]=VSAdeviceDriver(comNumber, commands.retBytes{command}, 1, motor_cmd', 1);    
       
       if motorResponse(1)~=3 && motorResponse(5)~=30 && motorFlag==0
          break
       end

       pause(0.5); 
       comRetry=comRetry+1;        
       if comRetry>5 
          if motorFlag==0
              motorreturn.status='FAILED. Motor error.'; 
              motorreturn.flag=-5;
          else
              motorreturn.status='FAILED. Communication Error.'; 
              motorreturn.flag=-6;              
          end
          return
       end            
    end                
 
    if motorResponse(4)==motor_cmd(4) && motorResponse(5)~=30 
        
       if command == 1
          timehex=[dec2hex(motorResponse(5),2); dec2hex(motorResponse(6),2); dec2hex(motorResponse(7),2); dec2hex(motorResponse(8),2)];
          motorreturn.estimatedTime=ieee2dec(timehex);      

       elseif command == 2          
           positionhex=[dec2hex(motorResponse(5),2); dec2hex(motorResponse(6),2); dec2hex(motorResponse(7),2); dec2hex(motorResponse(8),2)];
           motorreturn.position=ieee2dec(positionhex);          
     
           infoBlock=fliplr(dec2bin(motorResponse(9), 8));
          
           motorreturn.referenced      = str2double(infoBlock(1)); 
           motorreturn.inMotion        = str2double(infoBlock(2)); 
           motorreturn.warning         = str2double(infoBlock(4)); 
           motorreturn.error           = str2double(infoBlock(5));
           motorreturn.breakActive     = str2double(infoBlock(6));
           motorreturn.motionFinished  = str2double(infoBlock(7));
           motorreturn.targetReached   = str2double(infoBlock(8));          

           if  deviceconfig.invert{deviceid}
               motorreturn.position=deviceconfig.offset{deviceid}-motorreturn.position;
           end
       end      
        
       motorreturn.status='OK'; 
       motorreturn.flag=0;
       break   
    end  

    if cmdRetry>4       
       break
    end    
    cmdRetry=cmdRetry+1;   
end 



function r=ieee2dec(hexnumber)

w=([hexnumber(4), hexnumber(8), hexnumber(3), hexnumber(7), hexnumber(2), hexnumber(6), hexnumber(1), hexnumber(5)]); 

 x = '';
 for i=1:length(w)
     x = strcat(x,h2b(w(i)));
 end

z = 1;
if (iszeroieee754(x))
    r = 0;
    return;
end
if (isinfieee754(x))
    r = inf;
    return;
end
if (isninfieee754(x))
    r = -inf;
    return;
end

    eV = x(2:9);
    mV = x(10:32);
    b = 127;

if (x(1)=='0')
    s = 1;
else
    s = -1;
end

e = 0;
m = 1;
for i=1:length(eV)
    e = e + str2double(eV(i))*(2^(length(eV)-i));
end
for i=1:length(mV)
    m = m + str2double(mV(i))*(2^-i);
end
r = s*m*2^(e-b);

function y=iszeroieee754(x)
y=1;
for i=2:length(x)
    if (x(i)=='1')
        y = 0;
        break;
    end
end
return

function y=isinfieee754(x)
if (x(1)=='1')
    y=0;
    return
end
y=1;

    imax = 9;

for i=2:imax
    if (x(i)=='0')
        y = 0;
        break;
    end
end
return

function y=isninfieee754(x)
if (x(1)=='0')
    y=0;
    return
end
y=1;
imax = 9;
for i=2:imax
    if (x(i)=='0')
        y = 0;
        break;
    end
end
return

function x=h2b(w)
 X = ['0000';  % 0
      '0001';  % 1
      '0010';  % 2
      '0011';  % 3
      '0100';  % 4
      '0101';  % 5
      '0110';  % 6
      '0111';  % 7
      '1000';  % 8
      '1001';  % 9
      '1010';  % a
      '1011';  % b
      '1100';  % c
      '1101';  % d
      '1110';  % e
      '1111']; % f
  x = X(sscanf(w,'%x')+1,:);
  
  
  function x=dec2ieee(r)
y = sprintf('%tx',r);
x=([y(7:8); y(5:6); y(3:4); y(1:2)]);



%  o o o o o o     
%    o o o o o                                           
%    o o o o o    VariSphear VSA                            
%    o o o o o    Variable Spherical Microphone Array  
%    o o o                                            
%                                                    
%    /// ASAR PROJECT GROUP          
%
%    Cologne University of Applied Sciences         
%    Deutsche Telekom Laboratories, Berlin          
%    Technical University of Berlin                 
%    Westdeutscher Rundfunk, Cologne  
%
% 
%    !!! NOT FOR STANDALONE USE : crc16motor.m
%        THIS FUNCTION IS USED BY THE MOTION ROUTINES 
%
%    The Function calculates the required cyclic checksum for 
%    to communicate with Schunk Motion Modules (Schunk Motion Protocol). 
%     
%
%
%  Author        :  Benny Bernschtz, 09-DEC-2010 v1.0 (release) 
%  Last modified :  09-DEC-2010 v1.0 by ---


function allbytes = crc16motor(commands)

if exist('crcTable.mat')~=2

lookup = ['0000', 'C0C1', 'C181', '0140', 'C301', '03C0', '0280', 'C241',...
'C601', '06C0', '0780', 'C741', '0500', 'C5C1', 'C481', '0440',...
'CC01', '0CC0', '0D80', 'CD41', '0F00', 'CFC1', 'CE81', '0E40',...
'0A00', 'CAC1', 'CB81', '0B40', 'C901', '09C0', '0880', 'C841',...
'D801', '18C0', '1980', 'D941', '1B00', 'DBC1', 'DA81', '1A40',...
'1E00', 'DEC1', 'DF81', '1F40', 'DD01', '1DC0', '1C80', 'DC41',...
'1400', 'D4C1', 'D581', '1540', 'D701', '17C0', '1680', 'D641',...
'D201', '12C0', '1380', 'D341', '1100', 'D1C1', 'D081', '1040',...
'F001', '30C0', '3180', 'F141', '3300', 'F3C1', 'F281', '3240',...
'3600', 'F6C1', 'F781', '3740', 'F501', '35C0', '3480', 'F441',...
'3C00', 'FCC1', 'FD81', '3D40', 'FF01', '3FC0', '3E80', 'FE41',...
'FA01', '3AC0', '3B80', 'FB41', '3900', 'F9C1', 'F881', '3840',...
'2800', 'E8C1', 'E981', '2940', 'EB01', '2BC0', '2A80', 'EA41',...
'EE01', '2EC0', '2F80', 'EF41', '2D00', 'EDC1', 'EC81', '2C40',...
'E401', '24C0', '2580', 'E541', '2700', 'E7C1', 'E681', '2640',...
'2200', 'E2C1', 'E381', '2340', 'E101', '21C0', '2080', 'E041',...
'A001', '60C0', '6180', 'A141', '6300', 'A3C1', 'A281', '6240',...
'6600', 'A6C1', 'A781', '6740', 'A501', '65C0', '6480', 'A441',...
'6C00', 'ACC1', 'AD81', '6D40', 'AF01', '6FC0', '6E80', 'AE41',...
'AA01', '6AC0', '6B80', 'AB41', '6900', 'A9C1', 'A881', '6840',...
'7800', 'B8C1', 'B981', '7940', 'BB01', '7BC0', '7A80', 'BA41',...
'BE01', '7EC0', '7F80', 'BF41', '7D00', 'BDC1', 'BC81', '7C40',...
'B401', '74C0', '7580', 'B541', '7700', 'B7C1', 'B681', '7640',...
'7200', 'B2C1', 'B381', '7340', 'B101', '71C0', '7080', 'B041',...
'5000', '90C1', '9181', '5140', '9301', '53C0', '5280', '9241',...
'9601', '56C0', '5780', '9741', '5500', '95C1', '9481', '5440',...
'9C01', '5CC0', '5D80', '9D41', '5F00', '9FC1', '9E81', '5E40',...
'5A00', '9AC1', '9B81', '5B40', '9901', '59C0', '5880', '9841',...
'8801', '48C0', '4980', '8941', '4B00', '8BC1', '8A81', '4A40',...
'4E00', '8EC1', '8F81', '4F40', '8D01', '4DC0', '4C80', '8C41',...
'4400', '84C1', '8581', '4540', '8701', '47C0', '4680', '8641',...
'8201', '42C0', '4380', '8341', '4100', '81C1', '8081', '4040'];

for k=0:length(lookup)/4-1
    table(k+1)=hex2dec(lookup((k*4)+1:(k*4)+4));
end

for h=1:256
    jasuper(h,1:16)=bitget(table(h),1:16);
end
lookuptable=fliplr(jasuper);
save crcTable.mat lookuptable      
end    
    
%------------
    
load crcTable.mat lookuptable
numberofcommands=size(commands,1);
checksum(1, 1:16)=[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]; 

for n=2:numberofcommands+1 
    data=fliplr(bitget(hex2dec(commands(n-1,1:2)),1:16));      
    no1=bitand(checksum(n-1,1:16),[1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0]); 
    no2=fliplr(bitget(bin2dec(num2str(no1))/256,1:16));               
    no3=bitand(checksum(n-1,1:16),[0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1]); 
    no4=bitand(data,[0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1]);     
    indextablebin=bitxor(no3, no4);                         
    indextable=bin2dec(num2str(indextablebin))+1;           
    tableentry=lookuptable(indextable,1:16);                
    checksum(n, 1:16)=bitxor(no2,tableentry);               
end

finalchecksum=checksum(numberofcommands+1, 1:16);
hexchecksum=dec2hex(bin2dec(num2str(finalchecksum)));

if length(hexchecksum)==3
byteb=['0',hexchecksum(1)];
bytea=hexchecksum(2:3);   
elseif length(hexchecksum)==2
bytea=hexchecksum(1:2);
byteb='00';
elseif length(hexchecksum)==1
bytea=[0,hexchecksum(1)];
byteb='00';
else 
bytea=hexchecksum(3:4);
byteb=hexchecksum(1:2);

end
allbytes=[commands;bytea;byteb]; 
 


